
def nim_sum(num_list):
    """ 
    Calculates the nim sum of the elements in num_list

    Args:
        num_list ([int]): List of numbers

    Returns:
        total_sum: int
    """
    total_sum = 0
    for num in num_list:
        total_sum ^= num
    return total_sum



from .utils import nim_sum

class PerfectPlayer:
    """
    A PerfectPlayer choses an optimal action for the game nim.
    """
    def __init__(self) -> None:
        pass

    def action(self, environment):
        """
        Returns an action that leads to a new environment with a nim-sum of zero (if that's possible).
        This strategy would produce a perfect play for environments with two players.

        Args:
            environment: nim.Environment
                PerfectPlayer choses an action for this environment

        Returns:
            (pile,num_stones): (int,int)
                the location and number of stones removed from a heap
        """
        current_nim_sum = nim_sum(environment.heap)
        if current_nim_sum != 0:
            for heap_idx, current_stones in enumerate(environment.heap):
                wanted_stones = current_nim_sum ^ current_stones
                if wanted_stones <= current_stones:
                    return (heap_idx, current_stones - wanted_stones)
        
        return environment.random_action()
                       
class AlmostPerfectPlayer:
    """
    An AlmostPerfectPlayer choses an optimal action for the game nim, except when the current environment is a weakness_position.
    """
    def __init__(self, weakness_positions) -> None:
        """
        Args:
            weakness_positions: environment #CHECK
        """
        self.weakness_positions = weakness_positions
        self.perfect = PerfectPlayer()

    def action(self, environment):
        """
        Acts as a perfect player unless it is in a weakness_positions, in which case, it returns a subobtimal action.

        Args:
            environment: nim.Environment
                AlmostPerfectPlayer choses an action for this environment

        Returns:
            (pile,num_stones): (int,int)
                the location and number of stones removed from a heap
        """
        if environment.heap in self.weakness_positions:
            return environment.random_action()
        
        return self.perfect.action(environment)

class RandomPlayer:
    """
    A RandomPlayer choses a random valid actions for the game nim.
    """
    def __init__(self) -> None:
        pass

    def action(self, environment):
        """
        Determines a random action for a player to make.

        Args:
            environment: nim.Environment
                RandomPlayer choses an action for this environment

        Returns:
            (pile,num_stones): (int,int)
                the location and number of stones removed from a heap
        """
        return environment.random_action()


def random_rollout(environment):
    """
    A new environment is created based on a random action on the input environment.

    Args:
        environment: nim.Environment

    Returns:
        environment: nim.Environment
    """
    return environment.random_action()


from mcts.tree_node import TreeNode
from random import choice


class MctsPlayer:
    def __init__(self,
                 cache=dict(),
                 exploration_param=0.5,):
        self.exploration_param = exploration_param
        self.cache = cache

    def action(self, environment):
        
        turn=environment.turn()

        if environment not in self.cache:
            self.cache[environment]=TreeNode(environment, self.cache)
        
        current_node = self.cache.get(environment)
        
        if not current_node.is_expanded:
            current_node.expansion()

        action=choice(current_node.selection(self.exploration_param))

        next_node=current_node.children[action]

        if not next_node.is_expanded:
            next_node.expansion()
        
        return action









from random import choice
from math import inf, log, sqrt

from .utils import random_rollout


class TreeNode:
    """
    Executes the four steps in mcts: selection, expansion, simulation, and backpropagation for a specific node.
    """
    def __init__(self, environment, cache=dict()):
        """
        Initializes a node for montecarlo tree search.

        Args:
            environment: nim.Environment
            cache: dict
                takes a node and returns its corresponding tree_node.
        """
        self.environment = environment
        self.cache = cache
        if environment in cache:
            stored_node = cache.get(environment)
            self.num_visits = stored_node.num_visits
            self.agent_to_value = stored_node.agent_to_value
            self.is_expanded = stored_node.is_expanded
            self.children = stored_node.children
            self.parents = stored_node.parents
        else:
            self.num_visits = 0
            self.agent_to_value = [0] * self.environment.num_agents()
            self.is_expanded = False
            self.children = dict()
            self.parents = set()

        self.cache[environment] = self

    def __repr__(self) -> str:
        return f"Environment {self.environment}, Parent {self.parent}, Action {self.action} "

    def selection(self, exploration_parameter=0.5): # theoretically the exploration parameter is equal to sqrt(2)
        """
        This method returns a list of valid actions with higher ucb score (specific formula used to determine).

        Args:
            exploration_parameter: int
                keeps the balance between exploring new parts of the tree or exploiting current parts that have been explored; defaults to sqrt(2)

        Returns:
            best_actions: list
                contains actions that will lead to a desirable result for the current player.
        """
        
        total_visits = 0
        for child in self.children.values():
            total_visits += child.num_visits

        best_actions = []
        best_score = -inf
        current_player = self.environment.turn()
        for action, child in self.children.items():
            child_value = child.agent_to_value[current_player]
            child_visits = child.num_visits
            score = child_value / child_visits + sqrt( 
                exploration_parameter *
                log(total_visits) / child_visits)
            if score == best_score:
                best_actions.append(action)
                best_score = score
            elif score > best_score:
                best_actions.clear()
                best_actions.append(action)
                best_score = score

        return best_actions

    def simulation(self, exploration_param=0.5, rollout_strategy=random_rollout):
        """
        Runs a current environment until a terminal state and returns its value. 
        For expanded nodes, it chooses the next environment based on the selection function. 
        For not expanded nodes, it chooses the next environment using a rollout_strategy.
        The rollout_strategy takes an environment and returns an action. 
        The default rollout_strategy will return a random action from all possible actions.

        Args:
            rollout_strategy: environment
                takes an environment and returns an action. Defaults to random_rollout.

        Returns:
            simulation_value: list 
        """

        current_node = self
        while current_node.is_expanded and not current_node.environment.is_terminal():
            best_action = choice(current_node.selection(exploration_param))
            current_node = current_node.children[best_action]

        current_environment = current_node.environment
        while not (current_environment.is_terminal()):
            current_environment = current_environment.what_if(rollout_strategy(current_environment))
        
        simulation_value = [
            current_environment.value(k)
            for k in range(current_environment.num_agents())
        ]

        # current_node.backpropagation(simulation_value)
        return simulation_value

   
    def expansion(self):
        """
        Populates the children of a node. 
        """
        if self.is_expanded:
            return

        for action in self.environment.valid_actions():
            child_environment = self.environment.what_if(action)

            if child_environment in self.cache:
                child_node = self.cache.get(child_environment)
            else:
                child_node = TreeNode(child_environment, self.cache)

            child_node.parents.add(self.environment)
            child_node_value = child_node.simulation()
            child_node.backpropagation(child_node_value)
            self.children[action] = child_node

        self.is_expanded = True

    def backpropagation(self, value):
        """
        Propagates the value from a node to all of its ancestors. 
        
        Args:
            value: int
                the specifc value of the environment (1 for a win and -1 for a loss) 
        """
        for i in range(len(self.agent_to_value)):
            self.agent_to_value[i] += value[i]

        self.num_visits += 1

        for parent in self.parents:
            parent_node = self.cache.get(parent)
            parent_node.backpropagation(value)


from context import nim
from context import utils
from context import players


def test_perfect_player_vs_random_player():
    environment = nim.Environment([3, 4, 5], 0, 2)
    player_0 = nim.PerfectPlayer()
    player_1 = nim.RandomPlayer()

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == 1
    assert last_environment.value(1) == -1


def test_random_player_vs_perfect_player():
    environment = nim.Environment([3, 4, 5], 0, 2)
    player_0 = nim.RandomPlayer()
    player_1 = nim.PerfectPlayer()

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == -1
    assert last_environment.value(1) == 1

def test_perfect_player_vs_perfect_player():
    environment = nim.Environment([3, 4, 5], 0, 2)
    player_0 = nim.PerfectPlayer()
    player_1 = nim.PerfectPlayer()

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == 1
    assert last_environment.value(1) == -1

def test_almost_perfect_player_vs_perfect_player_1():
    environment = nim.Environment([3, 4, 5], 0, 2)
    player_0 = nim.AlmostPerfectPlayer([])      # The player has no weaknesses
    player_1 = nim.PerfectPlayer()

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == 1
    assert last_environment.value(1) == -1

def test_almost_perfect_player_vs_perfect_player_2():
    environment = nim.Environment([3, 4, 5], 0, 2)
    player_0 = nim.AlmostPerfectPlayer([[3, 4, 5]])      # The player has no weaknesses
    player_1 = nim.PerfectPlayer()

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == -1
    assert last_environment.value(1) == 1


def test_perfect_player_vs_mcts_player():
    environment = nim.Environment([2, 3], 0, 2)
    player_0 = players.MctsPlayer()
    player_1 = nim.PerfectPlayer()

    for _ in range(10000):
        log = utils.play(environment, [player_0, player_1])
        (last_environment, _, _) = log[-1]
        game_value = [last_environment.value(k) for k in range(last_environment.num_agents())]
        player_0.cache[last_environment].backpropagation(game_value)

    player_0.exploration_param=0

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == 1
    assert last_environment.value(1) == -1

def test_almost_perfect_player_vs_mcts_player():
    environment = nim.Environment([2, 3], 0, 2)
    player_0 = nim.AlmostPerfectPlayer([[2, 3]])
    player_1 = players.MctsPlayer()


    utils.play(environment, [player_1, player_0])
    for _ in range(1000):
        log = utils.play(environment, [player_0, player_1])
        (last_environment, _, _) = log[-1]
        game_value = [last_environment.value(k) for k in range(last_environment.num_agents())]
        player_1.cache[last_environment].backpropagation(game_value)
    
    player_1.exploration_param=0

    log = utils.play(environment, [player_0, player_1])
    (last_environment, _, _) = log[-1]

    assert last_environment.is_terminal() == True
    assert last_environment.value(0) == -1
    assert last_environment.value(1) == 1

def test_random_player_vs_mcts_player():
    environment = nim.Environment([2, 3], 0, 2)
    player_0 = players.MctsPlayer(exploration_param=0.5)
    player_1 = nim.RandomPlayer()
    player_2 = nim.PerfectPlayer()

    for _ in range(10000):
        log = utils.play(environment, [player_0, player_1])
        (last_environment, _, _) = log[-1]
        game_value = [last_environment.value(k) for k in range(last_environment.num_agents())]
        player_0.cache[last_environment].backpropagation(game_value)
    
    player_0.exploration_param=0
    
    for _ in range(10):
        log = utils.play(environment, [player_0, player_1])
        (last_environment, _, _) = log[-1]

        assert last_environment.is_terminal() == True
        assert last_environment.value(0) == 1
        assert last_environment.value(1) == -1





def play(environment, players):
    """
    This function looks at the current environment, asks the corresponding player to make a move, 
    and updates the  current environment accordingly. The function returns a log with
    (environment, current_player, action) for each move.

    Args:
        environment: nim.Environment
        players: [description] #CHECK

    Returns:
        log_moves (list):  #CHECK
    """
    log_moves = []
    while not environment.is_terminal():
        current_player = environment.turn()
        action = players[current_player].action(environment)
        environment = environment.what_if(action)
        moves = [(environment, current_player, action)]
        log_moves = log_moves + moves
    return log_moves


from setuptools import find_packages, setup

setup(
    name='mcts_kds',
    packages=find_packages(include=['mcts','nim','players','utils']),
    description='MCTS Library',
    version = 1.0,
    download_url = "https://github.com/MadisonM1600/mcts/archive/refs/tags/1.0.tar.gz",
    author='Madison Mussari, supervised by Pedro Rangel',
    tests_require=['pytest==4.4.1'],
    test_suite='test',
)



from context import nim

def test_state():
    environment = nim.Environment([3, 2, 0], 1, 3)
    expected_environment = ((2, 3), 1, 3)
    assert environment.state() == expected_environment

    environment = nim.Environment([2, 3, 1], 1, 3)
    expected_environment = ((1, 2, 3), 1, 3)
    assert environment.state() == expected_environment

    environment = nim.Environment([2, 3, 1], 3, 3)  # NOTE: Maybe it is better to throw an error here
    expected_environment = ((1, 2, 3), 0, 3)
    assert environment.state() == expected_environment


def test_is_terminal():
    environment = nim.Environment([3], 1, 3)
    assert environment.is_terminal() == True

    environment = nim.Environment([2, 3, 1], 1, 3)
    assert environment.is_terminal() == False

def test_value():
    environment = nim.Environment([2], 1, 3)
    assert environment.value(0) == -1
    assert environment.value(1) == 1
    assert environment.value(2) == -1
    assert environment.value(3) == None  # FIXME: Throw error

    environment = nim.Environment([2, 3, 1], 1, 3)
    assert environment.value(0) == None
    assert environment.value(1) == None
    assert environment.value(2) == None
    assert environment.value(3) == None  # FIXME: Throw error


def test_turn():
    environment = nim.Environment([2, 3, 1], 0, 3)
    assert environment.turn() == 0

    environment = nim.Environment([2, 3, 1], 1, 3)
    assert environment.turn() == 1

    environment = nim.Environment([2, 3, 1], 2, 3)
    assert environment.turn() == 2


def test_what_if():
    environment = nim.Environment([2, 3, 1], 0,
                                  3)  # current state ([1, 2, 3], 0, 3)
    environment = environment.what_if(
        (2, 1))  # Takes 1 stone of third heap (index 2)
    expected_state = ((1, 2, 2), 1, 3)
    assert expected_state == environment.state()

    environment = environment.what_if(
        (2, 1))  # Takes 1 stone of third heap (index 2)
    expected_state = ((1, 1, 2), 2, 3)
    assert expected_state == environment.state()

    environment = environment.what_if(
        (1, 1))  # Takes 1 stone of second heap (index 1)
    expected_state = ((1, 2), 0, 3)
    assert expected_state == environment.state()

    environment = environment.what_if((1, 3))  # Invalid move
    expected_state = ((1, 2), 0, 3)  # The environment does not change
    assert expected_state == environment.state()

    environment = environment.what_if(
        (1, 3))  # Takes 3 stone of second heap (index 1)
    expected_state = ((1, 2), 0, 3)  # Invalid move. Environment does not change.
    assert expected_state == environment.state()

    environment = environment.what_if(
        (0, 1))  # Takes 1 stone from first heap (index 0)
    expected_state = ((2,), 1, 3)
    assert expected_state == environment.state()

    environment = environment.what_if(
        (0, 1))  # The environment is final. Action not allowed
    expected_state = ((2,), 1, 3)  # State does not change
    assert expected_state == environment.state()

def test_valid_actions():
    environment = nim.Environment([2, 2, 3], 0, 3)
    expected_valid_actions = {(0, 1), (0, 2), (1, 1), (1, 2), (2, 1), (2, 2), (2, 3)}
    returned_valid_actions = {valid for valid in environment.valid_actions()}
    assert expected_valid_actions == returned_valid_actions


def test_num_agents():
    for num_players in range(1, 10):
        environment = nim.Environment([2, 2, 3], 0, num_players)
        assert environment.num_agents() == num_players


# Plays a full game of Nim
def test_simple_game():
    # Initial states
    environment = nim.Environment([2, 1, 3], 1, 3)

    assert environment.state() == ((1, 2, 3), 1, 3)
    assert environment.is_terminal() == False
    assert environment.turn() == 1

    valid_actions = {(0, 1), (1, 1), (1, 2), (2, 1), (2, 2), (2, 3)}
    environment_valid_actions = {x for x in environment.valid_actions()}
    print(environment.valid_actions())
    assert valid_actions == environment_valid_actions

    # First move, take 2 stones from heap 1 (second heap)
    environment = environment.what_if((1, 2))

    assert environment.state() == ((1, 3), 2, 3)
    assert environment.is_terminal() == False
    assert environment.turn() == 2

    valid_actions = {(0, 1), (1, 1), (1, 2), (1, 3)}
    environment_valid_actions = {x for x in environment.valid_actions()}
    assert valid_actions == environment_valid_actions

    # Checking that invalid moves don't change the environment
    environment = environment.what_if((0, 2))

    assert environment.state() == ((1, 3), 2, 3)
    assert environment.is_terminal() == False
    assert environment.turn() == 2

    valid_actions = {(0, 1), (1, 1), (1, 2), (1, 3)}
    environment_valid_actions = {x for x in environment.valid_actions()}
    assert valid_actions == environment_valid_actions

    # Second move, take 1 stone out of heap 0 (first heap)
    environment = environment.what_if((0, 1))
    assert environment.state() == ((3,), 0, 3) # FIXME this orginally was (3,1,[3])
    assert environment.is_terminal() == True
    assert environment.turn() == 0
    assert environment.value(0) == 1
    assert environment.value(1) == -1
    assert environment.value(2) == -1

    valid_actions = set()
    environment_valid_actions = {x for x in environment.valid_actions()}
    assert valid_actions == environment_valid_actions


def test_from_state():
    original_state = ((1, 2, 3), 1, 3)
    environment = nim.from_state(original_state)
    environment_state = environment.state()

    assert original_state == environment_state


def test_random_action():
    from math import sqrt
    from math import ceil

    environment = nim.Environment([3, 4, 5], 0, 2)

    actions_seen = dict()
    number_repetitions = 1000

    for _ in range(number_repetitions):
        current_action = environment.random_action()
        actions_seen[current_action] = actions_seen.get(current_action, 0) + 1
    
    total_number_of_actions = len([a for a in environment.valid_actions()])
    expected_number_of_actions = number_repetitions / total_number_of_actions
    prob = 1 / total_number_of_actions
    threshold = ceil(3 * sqrt(number_repetitions * prob * (1 - prob)))

    for repetions in actions_seen.values():
        assert abs(repetions - expected_number_of_actions) <= threshold

def test_eq():
    heap = [1,2,3]
    environment_1 = nim.Environment(heap,2,3)
    environment_2 = nim.Environment(heap,2,3)
    environment_3 = nim.Environment(heap,0,2)

    assert environment_1 == environment_2
    assert environment_1 != environment_3

def test_hash():
    heap = [1,2,3]
    environment_1 = nim.Environment(heap,2,3)
    environment_2 = nim.Environment(heap,2,3)
    environment_3 = nim.Environment(heap,0,2)

    assert hash(environment_1) == hash(environment_2)
    assert hash(environment_1) != hash(environment_3)


from context import nim
from context import mcts


def make_expanded_node(heap=[3, 2, 4], current_player=0, num_players=2):
    environment = nim.Environment(heap, current_player, num_players)
    parent_node = mcts.TreeNode(environment)

    for action in environment.valid_actions():
        child_environment = environment.what_if(action)
        child_node = mcts.TreeNode(child_environment, parent_node.cache)
        parent_node.children[action] = child_node
        parent_node.cache[child_environment] = child_node

    parent_node.is_expanded = True

    return parent_node


# test the functions in the TreeNode class
def test_selection_1():
    parent_node = make_expanded_node()

    actions = []
    for count, (action, child) in enumerate(parent_node.children.items(), 1):
        child.num_visits = 2 * count
        child.agent_to_value = [count, -count]
        parent_node.num_visits += 2 * count
        actions.append(action)

    # Since all children have the same score, and exploration parameter is different from zero, we should explore the least visited child.
    expected_selection = {actions[0]}
    selected = set(parent_node.selection())
    assert expected_selection == selected


def test_selection_2():
    parent_node = make_expanded_node()

    actions = []
    for count, (action, child) in enumerate(parent_node.children.items(), 1):
        child.num_visits = 2 * count
        child.agent_to_value = [-count, count]
        parent_node.num_visits += 2 * count
        actions.append(action)

    # Since all children have the same score, and exploreation parameter is different from zero, we should explore the least visited child.
    expected_selection = {actions[0]}
    selected = set(parent_node.selection())
    assert expected_selection == selected


def test_selection_3():
    parent_node = make_expanded_node()
    
    actions = []
    for count, (action, child) in enumerate(parent_node.children.items(), 1):
        child.num_visits = 2 * count
        child.agent_to_value = [count, -count]
        parent_node.num_visits += 2 * count
        actions.append(action)

    # Since all children have the same score, and exploration parameter is zero, all children are equally good.
    expected_selection = set(actions)
    selected = set(parent_node.selection(0))
    assert expected_selection == selected


def test_selection_4():
    parent_node = make_expanded_node()

    actions = []
    for count, (action, child) in enumerate(parent_node.children.items(), 1):
        child.num_visits = count
        child.agent_to_value = [1, -1]
        parent_node.num_visits += count
        actions.append(action)

    # With exploration parameter equal to zero, we should select the best score (the first child).
    expected_selection = {actions[0]}
    selected = set(parent_node.selection(0))
    assert expected_selection == selected


def test_selection_5():
    parent_node = make_expanded_node(current_player=1)
    
    actions = []
    for count, (action, child) in enumerate(parent_node.children.items(), 1):
        child.num_visits = count
        child.agent_to_value = [1, -1]
        parent_node.num_visits += count
        actions.append(action)

    # With exploration parameter equal to zero, we should select the best score (the last child).
    expected_selection = {actions[-1]}
    selected = set(parent_node.selection(0))
    assert expected_selection == selected


def test_expansion_1():
    environment = nim.Environment([2, 3, 1], 0, 2)
    root_node = mcts.TreeNode(environment)

    expected_children_before_expansion = set()
    real_children_before_expansion = set(root_node.children)

    assert expected_children_before_expansion == real_children_before_expansion
    assert root_node.is_expanded == False

    root_node.expansion()
    
    expected_expanded_node = make_expanded_node([2, 3, 1], 0, 2)

    for action, child_node in root_node.children.items():
        assert child_node.environment == expected_expanded_node.children[action].environment

    assert root_node.is_expanded == True

def test_expansion_2():
    environment = nim.Environment([2, 2, 2], 2, 3)
    root_node = mcts.TreeNode(environment)

    expected_children_before_expansion = set()
    real_children_before_expansion = set(root_node.children)

    assert expected_children_before_expansion == real_children_before_expansion
    assert root_node.is_expanded == False

    root_node.expansion()
    
    expected_expanded_node = make_expanded_node([2, 2, 2], 2, 3)

    for action, child_node in root_node.children.items():
        assert child_node.environment == expected_expanded_node.children[action].environment

    assert root_node.is_expanded == True


def test_backpropogation():
    environment = nim.Environment([2, 2, 2], 2, 3)
    node = mcts.TreeNode(environment)
    node.expansion()
    node.agent_to_value = [0, 0, 0]
    node.num_visits = 0
    
    num_visits = 0
    for _, child_node in node.children.items():
        child_node.backpropagation([1, 0, -1])
        num_visits += 1
    
    assert node.num_visits == num_visits
    assert node.agent_to_value == [num_visits, 0, - num_visits]


def test_simulation_1():
    # An environment that always leads to a win
    heap = [1] * 5
    environment = nim.Environment(heap, 0, 2)
    root_node = mcts.TreeNode(environment)

    expected_value = [1, -1]
    real_value = root_node.simulation()
    assert expected_value == real_value

def test_simulation_2():
    # An environment that always leads to a loss
    heap = [1] * 6
    environment = nim.Environment(heap, 0, 2)
    root_node = mcts.TreeNode(environment)

    expected_value = [-1, 1]
    real_value = root_node.simulation()
    assert expected_value == real_value




class HumanPlayer:
    """
    
    """
    def __init__(self, str_to_action):
        """
        Args:
            str_to_action ([type]): [description]
        """
        self.str_to_action = str_to_action
        pass

    def action(self, environment):
        """

        Args:
            environment: nim.Environment

        Returns:
            action(int,int): Describes the location and number of stones removed from a heap by the player
        """
        print("Current Environment\n")
        print(environment)
        input_str = input("Enter your move: ")
        action = self.str_to_action(input_str)
        return action



import random

class Environment:
    """
    Creates a new nim environment.
    """
    def __init__(self, heap=[10, 10, 10], current_player=0, num_players=2):
        """
        Args:
            heap: [int]
                describes the number of heaps and stones in each heap. Defaults to [10, 10, 10].
            current_player: int
                keeps track of which player's turn it is. Defaults to 0.
            num_players: int
                takes note of the number of players in the game. Defaults to 2.
        """

        self.heap = [h for h in sorted(heap) if h > 0]
        self.current_player = current_player if current_player < num_players else 0
        self.num_players = num_players

    
    def turn(self): 
        """
        Determines which player's turn it is. 

        Returns:
            current_player: int 
                the identifier for the current player
        """
        return self.current_player


    def valid_actions(self):
        """
        Finds the possible actions that could be taken by the player in the current environment. 

        Returns:
            valid_action_iterable: [(pile,num_stones)]
                a valid heap and number of stones that a player could take from that heap. 
        """
        if self.is_terminal():
            return iter([])
        
        return ((i, j) for i in range(len(self.heap)) for j in range(1, self.heap[i]+ 1))

    def random_action(self):
        """
        Produces a valid random action.

        Returns:
           (pile,num_stones): (int,int)
                the location and number of stones removed from a heap
        """
        if self.is_terminal():
            return None

        pile = random.choices(range(len(self.heap)), weights=self.heap)[0]
        pile_stones = self.heap[pile]
        num_stones = random.randrange(0,pile_stones) + 1

        # return random.choice([a for a in self.valid_actions()])

        return (pile,num_stones)


    def what_if(self, action): 
        """
        Produces game state/environment assuming that player moves "action". 

        Args:
            action (int,int): Describes the location and number of stones removed from a heap by the player

        Returns:
            environment: nim.Environment
                the state which results from taking action
        """

        if action[0] < 0 or action[0] > len(self.heap) - 1:
            return Environment(self.heap, self.current_player, self.num_players)
        
        new_value = self.heap[action[0]] - action[1]
        if new_value < 0 or self.is_terminal() or action[1] <= 0:
            return Environment(self.heap, self.current_player, self.num_players)
        
        new_heap = self.heap.copy()
        new_heap[action[0]] = new_value
        
        if new_heap[action[0]]==0:
            new_heap.pop(action[0])
    
        new_current_player = (self.current_player + 1) % self.num_players 
        environment = Environment(new_heap, new_current_player, self.num_players)
        return environment


    def is_terminal(self):
        """
        Checks if the current heap represents a terminal position. 

        Returns:
            Boolean: True or False
                True if the state is terminal, False if it is not
        """
        return len(self.heap) == 1

    
    def value(self, current_player):
        """
        Determines the value of a terminal environment (last node in a branch). If the environment is not terminal, it returns None

        Args:
            current_player (int): the identifier for the current player

        Returns:
            score: int
                the value for this environment (1 for a win and -1 for a loss)
                when the environment is not terminal, it returns None.
        """
        score = None
        if self.is_terminal() and (current_player < self.num_players):
            score = 1 if self.current_player == current_player else -1
        return score
    
    def num_agents(self):
        """
        Returns:
            num_players: int
                number of agents in the game
        """
        return self.num_players

    
    def state(self):
        """
        Describes the current envrionment of the game.

        Returns:
            state: list
               (heap, current_player, num_players)  
        """
        state = (tuple(self.heap), self.current_player, self.num_players)
        return state
    

    def __repr__(self) -> str:
        """
        Makes the current environment information into a string.

        Returns:
            environment_str: str
                describes heap, current_player, and num_player information
        """
        return f"Heap {self.heap}, Current Player {self.current_player}, Number of Players {self.num_players}"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Environment):
            return self.heap == o.heap and self.current_player == o.current_player and self.num_players == o.num_players 
        return False
    
    def __hash__(self) -> int:
        return hash(repr(self))


def from_state(state):
    """
    Creates a nim environment from the current state.

    Args:
        state (list): (heap, current_player, num_players) 

    Returns:
        Environment: list
            returns a list of states
    """
    return Environment(list(state[0]), state[1], state[2])


def str_to_action(action_str):
    """
    Converts a string into an action for Nim.

    Args:
        action_str (str):
            A string of the form "(pile,num_stones)"

    Returns:
        (int, int):
            ((pile,num_stones)), actions intended by action_str
    """
    return tuple(map(int, action_str.split(",")))

